/*
 * hwreset.c
 *
 * This tool power cycles (or powers off) the IPMI system.
 *    
 * Author:  Andy Cress  andrew.r.cress@intel.com
 * Copyright (c) 2002 Intel Corporation. 
 *
 * 01/08/02 Andy Cress - created
 * 01/31/02 Andy Cress - converted to use ipmi_cmd_ia, 
 *                       added more user options
 * 02/06/02 Andy Cress - added ipmi_cmd_va
 * 02/22/02 Andy Cress - added -s to reboot to service partition
 * 07/02/02 Andy Cress v1.3 added more Usage text
 * 08/02/02 Andy Cress v1.4 moved common ipmi_cmd() code to ipmicmd.c
 * 09/24/02 Andy Cress - stubbed in OS shutdown option
 * 01/29/03 Andy Cress v1.5 added MV OpenIPMI support
 * 04/08/03 Andy Cress v1.6 added OS shutdown option (-o)
 * 05/02/03 Andy Cress v1.7 leave console redir alone in SET_BOOT_OPTIONS
 */
/*M*
Copyright (c) 2002, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include "ipmicmd.h"
 
/*
 * Global variables 
 */
char   fdebug    = 0;
char * progver   = "1.6";
char * progname  = "hwreset";

static int IPMI_Reset(uchar bpower, uchar bootopt)
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	uchar completionCode;
	uchar inputData[6];
	int status;

        /* May need to GetSystemBootOptions first to fill these in. */
	inputData[0] = 0x05;  // param, 5 = boot flags
	inputData[1] = 0x80;  // valid flags
        if (bootopt == 1)     // set up for service partition boot 
	     inputData[2] = 0x10;  // boot to svc partition
	else inputData[2] = 0x00;  // normal boot
	inputData[3] = 0x48;  // verbose disply, bypass pswd, same console redir
	inputData[4] = 0x00;  // no overrides
	inputData[5] = 0x00;  // 
        status = ipmi_cmd(SET_BOOT_OPTIONS, inputData, 6, responseData,
                        &responseLength, &completionCode, fdebug);
	if (status != ACCESS_OK) {
            printf("Cmd %x ccode %x, resp[0] = %x, resp[1] =  %x\n",
		   SET_BOOT_OPTIONS,
                   completionCode, responseData[0], responseData[1]);
	   }

        /* 0 = power down, 1 = power up, 2 = power cycle, 3 = hard reset */
        /* 4 = NMI interrupt, 5 = soft shutdown OS via ACPI  */
	if (bpower > 5) bpower = 3;  /* if invalid, use default */
	inputData[0] = bpower;  // hard reset

        status = ipmi_cmd(CHASSIS_CTL, inputData, 1, responseData,
                        &responseLength, &completionCode, fdebug);
        if (fdebug) {
                printf("Cmd %x code %x, resp[0] = %x, resp[1] =  %x\n",
		       CHASSIS_CTL, completionCode, 
		       responseData[0], responseData[1]);
                }
	if (status == ACCESS_OK) {
		//successful, done
		return(0);
	}
        else {
           printf("Error: chassis completion code=%x\n", completionCode); 
	   return(-1);
	   }

}  /*end IPMI_Reset()*/

int
main(int argc, char **argv)
{
   int ret;
   char c;
   uchar breset;
   uchar bopt;
   uchar fshutdown = 0;

   // progname = argv[0];
   breset = 3;  /* hard reset as default */
   bopt = 0;    /* Boot Options default */
   printf("%s ver %s\n", progname,progver);

   while ( (c = getopt( argc, argv,"dconrsx?")) != EOF ) 
      switch(c) {
          case 'd': breset = 0;     break;  /* power down */
          // case 'u': breset = 1;  break;  /* power up */
          case 'c': breset = 2;     break;  /* power cycle */
          case 'o': fshutdown = 1;  break;  /* shutdown OS */
          case 'n': breset = 4;     break;  /* interrupt (NMI) */
          case 'r': breset = 3;     break;  /* hard reset */
          case 's': bopt   = 1;     break;  /* hard reset to svc part */
          case 'x': fdebug = 1;     break;  /* debug messages */
	  default:
                printf("Usage: %s [-c -d -i -n -r -s -x]\n", progname);
                printf(" where -c  power Cycles the system\n");
                printf("       -d  powers Down the system\n");
                printf("       -r  hard Resets the system\n");
                printf("       -n  sends NMI to the system\n");
                printf("       -o  soft-shutdown OS\n");
                printf("       -s  reboots to Service Partition\n");
                printf("       -x  show eXtra debug messages\n");
                exit(1);
      }

   if (fshutdown) {
	/* 
	 * Since we are user mode, we can't let it fully shut down. 
	 * IPMI can trigger this by emulating an overtemp event.
	 */
	// system("init 0");
	// sleep(10);  
         breset = 5;     /* soft shutdown option */
   }
   if (breset == 0 || breset == 5) printf("%s: powering down ...\n",progname);
   else       printf("%s: resetting ...\n",progname);
   ret = IPMI_Reset(breset,bopt);

   if (ret == 0) {
      printf("%s: IPMI_Reset ok\n",progname);
   }  /* if ok */
   exit (ret);
}  /* end main()*/

/* end hwreset.c */
