/*M*
//  PVCS:
//      $Workfile:   ipmicmd.c  $
//      $Revision:   1.0  $
//      $Modtime:   22 Jul 2002 08:51:14  $
//      $Author:   andrew.r.cress@intel.com  $  
//
//  Define the ipmi_cmd routine and supporting logic to execute IPMI 
//  commands via one of the supported IPMI drivers:
//     /dev/ipmi0  /dev/ipmi/0    = MontaVista OpenIPMI 
//     /dev/imb                   = Intel IMB ipmidrvr (comes with ISC)
//     /dev/ipmikcs /dev/ipmi/kcs = valinux driver by San Mehat
// 
//  08/05/02 ARC - created 
//  08/15/02 ARC - added decode_cc
//  10/24/02 ARC - made cmd param ushort to be more unique
//  01/29/03 ARC - added MontaVista OpenIPMI driver support
//  07/25/03 ARC - added serial-over-lan commands
//  07/30/03 ARC - added GetThresholds, fix for ipmi_cmd_raw,
//                 changed some error messages
//  09/04/03 ARC - added debug messages for fDriverTyp first time
 *M*/
/*----------------------------------------------------------------------*
The BSD License 

Copyright (c) 2002, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>

#include "ipmi_ioctls.h"   /* for ipmi_cmd_va() */

#include "ipmicmd.h"  

// special IMB defines
#define MAX_NO_OF_RETRIES       3
#define DEFAULT_TIMEOUT       (1000 * 1000)
#define SEL_TIMEOUT            400L 

// Other IPMI values
#define PUBLIC_BUS      0
#define PRIVATE_BUS  0x03
#define BMC_SA       0x20
#define BMC_LUN         0

#define NCMDS   51
struct {
 ushort cmdtyp;
 uchar sa;
 uchar bus;
 uchar netfn;
 uchar lun;
 uchar len;  /*length of request data (FYI, but not used here) */
} ipmi_cmds[NCMDS] = { 
 /*empty,temp*/ 0, BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0,
 GET_SEL_ENTRY,    BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 6,
 GET_SEL_INFO,     BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 0,
 RESERVE_SEL,      BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 0,
 CLEAR_SEL,        BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 6,
 GET_SEL_TIME,     BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 6,
 GET_LAN_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 4,
 SET_LAN_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 4,
 GET_SER_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 4,
 SET_SER_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 4, /*5?*/
 SET_SER_MUX,      BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 2,
 GET_PEF_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 3,
 SET_PEF_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 3,
//  SET_PEF_ENABLE,   BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 4, /*old*/
 GET_DEVICE_SDR,   BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 6,
 GET_DEVSDR_INFO,  BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 0,
 RESERVE_DEVSDR_REP,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 0,
 GET_SENSOR_READING,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 1,
 GET_SENSOR_READING_FACTORS,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 2,
 GET_SENSOR_TYPE,  BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 1,
 GET_SENSOR_THRESHOLD,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 1,
 SET_SENSOR_THRESHOLD,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 8,
 GET_SENSOR_HYSTERESIS,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 2,
 SET_SENSOR_HYSTERESIS,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 4,
 GET_SDR,          BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 6,
 GET_SDR_REPINFO,  BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 0,
 RESERVE_SDR_REP,  BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 0,
 GET_FRU_INV_AREA, BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 1,
 READ_FRU_DATA,    BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 4,
 WRITE_FRU_DATA,   BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 20, /*3+N(17)*/
 GET_DEVICE_ID,    BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0,
 SET_USER_ACCESS,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 4,
 GET_USER_ACCESS,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 2,
 GET_USER_NAME,    BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 1,
 SET_USER_NAME,    BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 17,
 SET_USER_PASSWORD,BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 18,
 CHASSIS_CTL,      BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 1,
 CHASSIS_IDENTIFY, BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 1,
 GET_POWERON_HOURS,BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 0,
 MASTER_WRITE_READ,BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 4, /*or 3*/
 GET_SYSTEM_GUID,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0,
 WATCHDOG_GET,     BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0,
 WATCHDOG_SET,     BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 6,
 WATCHDOG_RESET,   BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0,
 SET_BOOT_OPTIONS, BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 2,
 GET_BOOT_OPTIONS, BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 3,
 ACTIVATE_SOL,     BMC_SA, PUBLIC_BUS, NETFN_SOL,   BMC_LUN, 0,
 SET_SOL_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_SOL,   BMC_LUN, 3,
 GET_SOL_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_SOL,   BMC_LUN, 4,
 GET_CHANNEL_INFO, BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 1,
 SET_CHANNEL_ACC,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 2,
 GET_CHANNEL_ACC,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 3 };

/* Special stuff for ipmi_cmd_ia() IMB functions */

// Request structure provided to SendTimedImbpRequest()
#pragma pack(1)
typedef struct {
        uchar   cmdType;        // IMB command
        uchar   rsSa;           // command destination address
        uchar   busType;        // not used
        uchar   netFn;          // IMB command class (network function)
        uchar   rsLun;          // subsystem on destination
        uchar * data;           // command body
        int     dataLength;     // body size
} IMBPREQUESTDATA;
#pragma pack() 

extern int ipmicmd_mv(uchar cmd, uchar netfn, uchar lun, uchar *pdata, 
			uchar sdata, uchar *presp, int sresp, int *rlen);
extern int ipmi_open_mv(void);
extern int ipmi_close_mv(void);
extern int ipmi_open_ia(void);
extern int ipmi_close_ia(void);
extern uint SendTimedImbpRequest (   /* defined in imbapi.c */
        IMBPREQUESTDATA *reqPtr,         // request info and data
        int              timeOut,        // how long to wait, in mSec units
        uchar           *respDataPtr,    // where to put response data
        int             *respDataLen,    // how much response data there is
        uchar           *completionCode  // request status from dest controller
        );

/* Global Data */
int fDriverTyp = 0;   /* 1 = IMB driver, 2 = VA driver, 3 = MV driver */
FILE *fperr = NULL;
FILE *fpdbg = NULL;

int ipmi_cmd_ia(ushort cmd, uchar *pdata, uchar sdata, uchar *presp, 
		int *sresp, uchar *pcc, char fdebugcmd)
{
   IMBPREQUESTDATA requestData;
   int status = 0;
   struct stat     stbuf;
   char *imbDev;
   int i;

   imbDev = "/dev/imb";

   for (i = 0; i < NCMDS; i++) {
      if (ipmi_cmds[i].cmdtyp == cmd) break;
   }
   if (i >= NCMDS) {
	fprintf(fperr,"ipmi_cmd_ia: Unknown command %x\n",cmd);
	return(-1); 
	}
   if (cmd >= CMDMASK) cmd = cmd & CMDMASK;  /* unmask it */

   requestData.cmdType	= cmd;
   requestData.rsSa	= ipmi_cmds[i].sa;
   requestData.busType	= ipmi_cmds[i].bus;
   requestData.netFn	= ipmi_cmds[i].netfn;
   requestData.rsLun	= ipmi_cmds[i].lun;
   requestData.dataLength = sdata;
   requestData.data       = pdata; 

   if (fdebugcmd) {
          uchar * pc; int sz;
          sz = sizeof(IMBPREQUESTDATA);
          pc = (uchar *)&requestData.cmdType;
          fprintf(fpdbg,"ipmi_cmd_ia: request (len=%d): ",sz);
          for (i = 0; i < sz; i++) fprintf(fpdbg,"%02x ",pc[i]);
          fprintf(fpdbg,"\n");
          pc = requestData.data;
          sz = requestData.dataLength;
          fprintf(fpdbg,"  req.data=%p, dlen=%d: ", pc, sz);
          for (i = 0; i < sz; i++) fprintf(fpdbg,"%02x ",pc[i]);
          fprintf(fpdbg,"\n");
        }

   if (stat(imbDev, &stbuf) != -1)  {
	   memset(presp, 0, *sresp);  // note that sresp must be pre-set 
           for ( i =0 ; i < MAX_NO_OF_RETRIES; i++) 
	   {
		if((status =SendTimedImbpRequest(&requestData, DEFAULT_TIMEOUT, presp, sresp, pcc)) == ACCESS_OK ) {
			break;
			}
		if (fdebugcmd)   // only gets here if error
	          fprintf(fpdbg,"ipmi_cmd_ia: sendImbRequest status=%x, ccode=%x\n",
                            (uint)status, *pcc);
	   } 
   } else {
	   fprintf(fperr,"ipmi_cmd_ia: No IMB driver found (%s)\n",imbDev);
	   return(-2);
   }

    if (fdebugcmd) {
	    fprintf(fpdbg,"ipmi_cmd_ia: sendImbRequest status=%x, ccode=%x\n", 
                      (uint)status, *pcc);
            if (status == ACCESS_OK) {
               uchar * pc; int sz;
               sz = *sresp;
               pc = (uchar *)presp;
               fprintf(fpdbg,"ipmi_cmd_ia: response (len=%d): ",sz);
               for (i = 0; i < sz; i++) fprintf(fpdbg,"%02x ",pc[i]);
               fprintf(fpdbg,"\n");
            }
        }

    return(status);
} /* end ipmi_cmd_ia() */

int ipmi_cmd_mv(ushort cmd, uchar *pdata, uchar sdata, uchar *presp, 
		int *sresp, uchar *pcc, char fdebugcmd)
{
    uchar  buf[MAX_BUFFER_SIZE];
    int rc, i, buflen;
    uchar   cc;
    int rlen = 0;

    for (i = 0; i < NCMDS; i++) {
       if (ipmi_cmds[i].cmdtyp == cmd) break;
    }
    if (i >= NCMDS) {
	fprintf(fperr, "ipmi_cmd_mv: Unknown command %x\n",cmd);
	return(-1); 
	}
    if (cmd >= CMDMASK) cmd = cmd & CMDMASK;  /* unmask it */

    if (fdebugcmd) fprintf(fpdbg,
			"mv cmd=%02x netfn=%02x lun=%02x sdata=%d sresp=%d\n",
			cmd,ipmi_cmds[i].netfn,ipmi_cmds[i].lun,sdata,
			*sresp); 
    rc = ipmicmd_mv(cmd,ipmi_cmds[i].netfn,ipmi_cmds[i].lun,
			pdata,sdata,buf,(*sresp + 1),&rlen);
    // if (rc == -1) printf("ipmi_cmd_mv: cannot open /dev/ipmi0\n");
    cc = buf[0];
    if (fdebugcmd) {
	    fprintf(fpdbg,"ipmi_cmd_mv: ipmicmd_mv status=%x, ccode=%x\n", 
                      (uint)rc, cc);
            if (rc == ACCESS_OK) {
               uchar * pc; int sz;
               sz = rlen;
               pc = (uchar *)buf;
               fprintf(fpdbg,"ipmi_cmd_mv: response (len=%d): ",sz);
               for (i = 0; i < sz; i++) fprintf(fpdbg,"%02x ",pc[i]);
               fprintf(fpdbg,"\n");
            }
        }
    *pcc = cc;
    if (rlen > 0) {
       /* copy data, except first byte */
       rlen -= 1;
       if (rlen > *sresp) rlen = *sresp;
       memcpy(presp,&buf[1],rlen);
    }
    *sresp = rlen;

    return(rc);
}  /*end ipmi_cmd_mv*/

/* #define   IOCTL_IPMI_XFER         0x01  (in ipmi_ioctls.h) */
int fd_va = -1;

int ipmi_open_va(void)
{
    char *ipmiDev;
    int fd;
    ipmiDev = "/dev/ipmikcs";  
    if ((fd = open(ipmiDev, O_RDWR))<0) {
    	/* fprintf(fperr,"ipmi_cmd_va: cannot open %s\n", ipmiDev); */
        ipmiDev = "/dev/ipmi/kcs";  /* some use this name instead */
	}
    if ((fd = open(ipmiDev, O_RDWR))<0) {
    	/* fprintf(fperr,"ipmi_cmd_va: cannot open %s\n", ipmiDev); */
	return(-2);
	}
    fd_va = fd;
    return(0);
}

int ipmi_close_va(void)
{
    int rc = 0;
    if (fd_va != 0) rc = close(fd_va);
    return(rc);
}

int ipmi_cmd_va(ushort cmd, uchar *pdata, uchar sdata, uchar *presp, 
		int *sresp, uchar *pcc, char fdebugcmd)
{
    IPMI_XFER                       xfer;
    IPMI_XFER                       *xferp = &xfer;
    int fd, rc, i;
    uchar   cc;

    for (i = 0; i < NCMDS; i++) {
       if (ipmi_cmds[i].cmdtyp == cmd) break;
    }
    if (i >= NCMDS) {
	fprintf(fperr, "ipmi_cmd_va: Unknown command %x\n",cmd);
	return(-1); 
	}
    if (cmd >= CMDMASK) cmd = cmd & CMDMASK;  /* unmask it */
 
    if (ipmi_open_va() != 0) return(-2);
    fd = fd_va;

    INIT_XFER(xferp);
    SET_REQUEST_LUN(xferp, ipmi_cmds[i].lun);
    SET_REQUEST_NETFN(xferp, ipmi_cmds[i].netfn);
    SET_REQUEST_CMD(xferp, cmd);
    if (sdata > 0) 
	SET_REQUEST_DATA(xferp, (uchar *) pdata, sdata);
    if ((rc = ioctl(fd, IOCTL_IPMI_XFER, (void *) &xfer))<0) {
                fprintf(fperr,"ipmi_cmd_va: failed (%m)\n");
                return(-1);
               }
    GET_RESPONSE_CC(xferp, cc);
    *pcc = cc;
    if (cc != 00) {
	fprintf(fperr,"ipmi_cmd_va: completionCode = %x\n",cc);
	return(-1);
	}
    else {
	GET_RESPONSE_DATA(xferp, presp);
	}

    if (fdebugcmd) {
	    fprintf(fpdbg,"ipmi_cmd_va: ioctl status=%x, ccode=%x\n",(uint)rc, cc);
            if (rc == 0) {
               uchar * pc; int sz;
               sz = *sresp;
               pc = (uchar *)presp;
               fprintf(fpdbg,"ipmi_cmd_va: response (len=%d): ",sz);
               for (i = 0; i < sz; i++) fprintf(fpdbg,"%02x ",pc[i]);
               fprintf(fpdbg,"\n");
            }
        }
    ipmi_close_va();
    return(rc);
} /* end ipmi_cmd_va() */

/* From IPMI v1.5 spec, Table 5-2 Completion Codes */
#define NUMCC    24
struct {
	uchar code;
	char *mesg;
	}  cc_mesg[NUMCC] = { 
0x00, "Command completed successfully",
0xC0, "Node Busy",
0xC1, "Invalid Command",
0xC2, "Command invalid for given LUN",
0xC3, "Timeout while processing command",
0xC4, "Out of space",
0xC5, "Invalid Reservarion ID, or cancelled",
0xC6, "Request data truncated",
0xC7, "Request data length invalid",
0xC8, "Request data field length limit exceeded",
0xC9, "Parameter out of range",
0xCA, "Cannot return requested number of data bytes",
0xCB, "Requested sensor, data, or record not present",
0xCC, "Invalid data field in request",
0xCD, "Command illegal for this sensor/record type",
0xCE, "Command response could not be provided",
0xCF, "Cannot execute duplicated request",
0xD0, "SDR Repository in update mode, no response",
0xD1, "Device in firmware update mode, no response",
0xD2, "BMC initialization in progress, no response",
0xD3, "Destination unavailable",
0xD4, "Cannot execute command. Insufficient privilege level",
0xD5, "Cannot execute command. Request parameters not supported",
0xFF, "Unspecified error"
};

char * decode_cc(uchar cc)
{
   char *pmsg;
   int i;
   for (i = 0; i < NUMCC; i++) {
	if (cc == cc_mesg[i].code) break;
   }
   pmsg = cc_mesg[i].mesg;
   return(pmsg);
}

/* 
 * ipmi_cmd_raw()
 *
 * This routine can be used to dynamically invoke commands that are not
 * already defined in the ipmi_cmds array by filling in those values 
 * in the empty slot in that array (at index 0).
 * If this app is multi-threaded, this won't be sufficient, since this 
 * empty array slot could be overwritten.  This routine is only 
 * intended for occasional use.
 * It is otherwise similar to ipmi_cmd().
 */
int ipmi_cmd_raw(uchar cmd, uchar netfn, uchar sa, uchar bus, uchar lun,
		uchar *pdata, uchar sdata, uchar *presp,
                int *sresp, uchar *pcc, char fdebugcmd)
{
    int rc;
    ushort icmd;

    fperr = stderr;
    fpdbg = stdout;

    if (fDriverTyp == 0) {   /*first time, so find which one */
	rc = ipmi_open_mv(); 
	if (rc == ACCESS_OK) {
		fDriverTyp = 3;
		ipmi_close_mv();
	} else {
         rc = ipmi_open_ia();
         if (rc == ACCESS_OK) 
		fDriverTyp = 1;
	 else {
           rc = ipmi_open_va();
           if (rc == ACCESS_OK) {
		fDriverTyp = 2;
		ipmi_close_va();
		}
	   else fprintf(fperr,
			"Cannot open any IPMI driver: /dev/imb, /dev/ipmi0, "
			"\n\t/dev/ipmi/0, /dev/ipmikcs, or /dev/ipmi/kcs.\n");
	 }
	}
	if (fdebugcmd) fprintf(fperr,"Driver type %d, rc = %d\n",fDriverTyp,rc);
    }  /*endif first time*/

    /* Set up the temporary ipmi_cmds array structure.  */
    icmd = cmd | (netfn << 8);
    ipmi_cmds[0].cmdtyp = icmd;
    ipmi_cmds[0].sa    = sa;
    ipmi_cmds[0].bus   = bus;
    ipmi_cmds[0].netfn = netfn;
    ipmi_cmds[0].lun   = lun;
    ipmi_cmds[0].len   = sdata;  /*length of request data*/
    *pcc = 0;

    if (fDriverTyp == 2) {
       rc = ipmi_cmd_va(icmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
    } else if (fDriverTyp == 1) {
       rc = ipmi_cmd_ia(icmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
    } else if (fDriverTyp == 3) {
       rc = ipmi_cmd_mv(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
    } else {     /* no IPMI driver found */
       rc = -2;
    }   
    if ((rc == ACCESS_OK) && (*pcc != 0) && fdebugcmd)
          fprintf(fpdbg,"raw ccode %x: %s\n",*pcc,decode_cc(*pcc));

    return(rc);
}

/* 
 * ipmi_cmd()
 *
 * This is the externally exposed subroutine for commands that
 * are defined in the ipmi_cmds array above.
 * It invokes whichever driver-specific routine is needed (va or ia).
 */
int ipmi_cmd(ushort cmd, uchar *pdata, uchar sdata, uchar *presp,
                int *sresp, uchar *pcc, char fdebugcmd)
{
    int rc;

    fperr = stderr;
    fpdbg = stdout;
#ifdef DEBUG
       { // if (fdebugcmd) {
		fprintf(fpdbg,"stdin %x, stdout %x, stderr %x\n",
			stdin,stdout,stderr);
	}
#endif
    if (fDriverTyp == 0) {   /*first time, so find which one */
	rc = ipmi_open_mv(); 
	if (rc == ACCESS_OK) {
		fDriverTyp = 3;
		ipmi_close_mv();
	} else {
         rc = ipmi_open_ia();
         if (rc == ACCESS_OK) 
		fDriverTyp = 1;
	 else {
           rc = ipmi_open_va();
           if (rc == ACCESS_OK) {
		fDriverTyp = 2;
		ipmi_close_va();
		}
	   else fprintf(fperr,
			"Cannot open any IPMI driver: /dev/imb, /dev/ipmi0, "
			"\n\t/dev/ipmi/0, /dev/ipmikcs, or /dev/ipmi/kcs.\n");
	 }
	}
	if (fdebugcmd) fprintf(fperr,"Driver type %d, rc = %d\n",fDriverTyp,rc);
    }  /*endif first time*/

    if (fDriverTyp == 2) {
       rc = ipmi_cmd_va(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
    } else if (fDriverTyp == 1) {
       rc = ipmi_cmd_ia(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
    } else if (fDriverTyp == 3) {
       rc = ipmi_cmd_mv(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
    } else {    /*no ipmi driver*/
	rc = -2;
    }
    if ((rc == ACCESS_OK) && (*pcc != 0) && fdebugcmd)
          fprintf(fpdbg,"ccode %x: %s\n",*pcc,decode_cc(*pcc));

    return(rc);
}

int ipmi_open(void)
{
   int rc = 0;
   fperr = stderr;
   fpdbg = stdout;

   /* first time, so try each */
   rc = ipmi_open_mv();
   if (rc == 0) fDriverTyp = 3;
   else {
	rc = ipmi_open_ia();
	if (rc == 0) fDriverTyp = 1;
	else {
		rc = ipmi_open_va(); 
		if (rc == 0) fDriverTyp = 2;
	}
   }
   if (rc != 0) fprintf(fperr,"ipmi_open rc = %d\n",rc);
   return (rc);
}

int ipmi_close(void)
{
   int rc = 0;
   switch (fDriverTyp)
   {
	case 1: rc = ipmi_close_ia(); break;
	case 2: rc = ipmi_close_va(); break;
	case 3: rc = ipmi_close_mv(); break;
	default:  break;
   }  /*end switch*/
   return (rc);
}

int ipmi_getdeviceid(uchar *presp, int sresp)
{
	int rc; uchar cc;
	/* check that presp is big enough (default is 15 bytes for Langley)*/
	if (sresp < 15) return(-3);
        rc = ipmi_cmd(GET_DEVICE_ID, NULL, 0, presp,&sresp, &cc, 0);
	if (rc != ACCESS_OK) return(rc);
	if (cc != 0) return(cc);
	return(ACCESS_OK);   /* success */
}

