/*
 * sensor.c
 *
 * This tool reads the SDR records to return sensor information. 
 * It can use either the Intel /dev/imb driver or VALinux /dev/ipmikcs.
 *
 * Author:  andrew.r.cress@intel.com
 * Copyright (c) 2002-2003 Intel Corporation. 
 *
 * 07/25/02 Andy Cress created
 * 10/09/02 Andy Cress v1.1 added decodeValue routine 
 * 10/11/02 Andy Cress v1.2 added expon routine 
 * 10/30/02 Andy Cress v1.3 added SDR types 08 & 14
 * 12/04/02 Andy Cress v1.4 changed dstatus descriptions
 * 01/29/03 Andy Cress v1.5 added MV OpenIPMI driver support
 *          Hannes Schulz <schulz@schwaar.com>
 *                1) correct raw readings to floatval
 *                2) allow extra SDR bytes returned from HP netserver 1000r
 *          Guo Min <guo.min@intel.com>
 *                add -l option for simpler list display
 * 02/25/03 Andy Cress v1.6 misc cleanup
 * 05/02/03 Andy Cress v1.7 add PowerOnHours
 * 07/28/03 Andy Cress v1.8 added -t option for threshold values,
 *                          added sample Discovery routine (unfinished),
 *                          added ipmi_getdeviceid for completeness.
 * 09/05/03 Andy Cress v1.9 show SDR OEM subtypes, 
 *                          fix GetSDR multi-part get for OEM SDRs
 *                          stop if SDR Repository is empty
 * 09/23/03 Andy Cress v1.10 Add options to set thresholds
 * 10/14/03 Andy Cress v1.11 Fixed sdr offset for ShowThreshold values
 * 01/15/04 Andy Cress v1.12 Fixed SetThreshold to set hysteresis, 
 *                           Fixed sens_cap testing in ShowThresh(Full)
 */
/*M*
Copyright (c) 2002, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>   // for:  double pow(double x, double y);
#include <string.h>
#include <getopt.h>
#include "ipmicmd.h"

/************************
 *  Global Data
 ************************/
char progname[] = "sensor";
char progver[] = "1.12";
char fdebug = 0;
int flist  = 0;
int fshowthr  = 0;
char tmpstr[20];  /* temp string */
int   fsetthresh = 0;
uchar sensor_num = 0xff;
uchar sensor_hi = 0xff;
uchar sensor_lo = 0xff;

/* sensor_dstatus
 * TODO: This needs further work.  The meaning depends on other factors. 
 */
static const char *sensor_dstatus[] = {
/* 0 00h */ "OK  ",
/* 1 01h */ "Warn-lo",   // "Warning-lo",
/* 2 02h */ "Crit-lo",   // "Critical-lo",
/* 3 04h */ "BelowCrit", // "BelowCrit-lo",
/* 4 08h */ "Warn-hi",   // "Warning-hi",
/* 5 10h */ "Crit-hi",   // "Critical-hi",
/* 6 20h */ "AboveCrit", // "AboveCrit-hi",
/* 7 40h */ "state 7 ",
/* 8 80h */ "OK* ",
/* 9 HSC */ "present ",
/*10 HSC */ "missing ",
/*11 HSC */ "ready   ",
/*12 HSC */ "faulty  "
};

static const char *sensor_types[] = {
/* 00h */ "reserved",
/* 01h */ "Temperature",
/* 02h */ "Voltage",
/* 03h */ "Current",
/* 04h */ "Fan",
/* 05h */ "Platform Chassis Intrusion",
/* 06h */ "Platform Security Violation",
/* 07h */ "Processor",
/* 08h */ "Power Supply",
/* 09h */ "Power Unit",
/* 0Ah */ "Cooling Device",
/* 0Bh */ "FRU Sensor",
/* 0Ch */ "Memory",
/* 0Dh */ "Drive Slot",
/* 0Eh */ "POST Memory Resize",
/* 0Fh */ "System Firmware",
/* 10h */ "SEL Disabled",
/* 11h */ "Watchdog 1",
/* 12h */ "System Event",          /* offset 0,1,2 */
/* 13h */ "Critical Interrupt",    /* offset 0,1,2 */
/* 14h */ "Button",                /* offset 0,1,2 */
/* 15h */ "Board",
/* 16h */ "Microcontroller",
/* 17h */ "Add-in Card",
/* 18h */ "Chassis",
/* 19h */ "Chip Set",
/* 1Ah */ "Other FRU",
/* 1Bh */ "Cable / Interconnect",
/* 1Ch */ "Terminator",
/* 1Dh */ "System Boot Initiated",
/* 1Eh */ "Boot Error",
/* 1Fh */ "OS Boot",
/* 20h */ "OS Critical Stop",
/* 21h */ "Slot / Connector",
/* 22h */ "ACPI Power State",
/* 23h */ "Watchdog 2",
/* 24h */ "Platform Alert",
/* 25h */ "Entity Presence",
/* 26h */ "Monitor ASIC",
/* 27h */ "LAN",
/* 28h */ "Management Subsystem Health",
/* 29h */ "Battery",
};

#define NUNITS  21
static const char *unit_types[] = {
/* 00 */ "unspecified",
/* 01 */ "degrees C",
/* 02 */ "degrees F",
/* 03 */ "degrees K",
/* 04 */ "Volts",
/* 05 */ "Amps",
/* 06 */ "Watts",
/* 07 */ "Joules",
/* 08 */ "Coulombs",
/* 09 */ "VA",
/* 10 */ "Nits",
/* 11 */ "lumen",
/* 12 */ "lux",
/* 13 */ "Candela",
/* 14 */ "kPa",
/* 15 */ "PSI",
/* 16 */ "Newton",
/* 17 */ "CFM",
/* 18 */ "RPM",
/* 19 */ "Hz",
/* 20 */ "microsecond",
/* 21 */ "millisecond",
};


#define NGDESC   6
struct {
 uchar g_id;
 const char desc[10];
} gen_desc[NGDESC] = {
  0x00, "IPMB",
  0x03, "BIOS",
  0x20, "BMC ",
  0x21, "SMI ",
  0x28, "CHAS",
  0xc0, "HSC "
};

typedef struct {
  ushort recid;
  uchar  sdrver;  /*usu. 0x51 = v1.5*/
  uchar  rectype; /* 01, 02, 11, 12, c0 */
  uchar  reclen;
  uchar  sens_ownid;
  uchar  sens_ownlun;
  uchar  sens_num;
  uchar  entity_id;
  uchar  entity_inst;
  uchar  sens_init; 
  uchar  sens_capab; 
  uchar  sens_type; 
  uchar  data1[7];
  uchar  sens_units;
  uchar  sens_base;
  uchar  sens_mod;
  uchar  linear;
  uchar  m;
  uchar  m_t;
  uchar  b;
  uchar  b_a;
  uchar  a_ax;
  uchar  rx_bx;
  uchar  flags;
  uchar  nom_reading;
  uchar  norm_max;
  uchar  norm_min;
  uchar  sens_max_reading;
  uchar  sens_min_reading;
  uchar  data3[11];
  uchar  id_strlen;
  uchar  id_string[16];
  } SDR01REC;

typedef struct {
  ushort recid;
  uchar  sdrver;  /*usu. 0x51 = v1.5*/
  uchar  rectype; /* 01, 02, 11, 12, c0 */
  uchar  reclen;
  uchar  sens_ownid;
  uchar  sens_ownlun;
  uchar  sens_num;
  uchar  entity_id;
  uchar  entity_inst;
  uchar  sens_init; 
  uchar  sens_capab; 
  uchar  sens_type; 
  uchar  data1[7];
  uchar  sens_units;
  uchar  sens_base;
  uchar  sens_mod;
  uchar  data2[7];
  uchar  id_strlen;
  uchar  id_string[16];
  } SDR02REC;

typedef struct {  /* 0x08 = Entity Association record */
  ushort recid;
  uchar  sdrver;  /*usu. 0x51 = v1.5*/
  uchar  rectype; /* 01, 02, 11, 12, c0 */
  uchar  reclen;
  uchar  contid;
  uchar  continst;
  uchar  flags;
  uchar  edata[8];
  } SDR08REC;

typedef struct {  /*0x14 = BMC Message Channel Info record */
  ushort recid;
  uchar  sdrver;  /*usu. 0x51 = v1.5*/
  uchar  rectype; /* 01, 02, 11, 12, c0 */
  uchar  reclen;
  uchar  mdata[8];
  uchar  mint;
  uchar  eint;
  uchar  rsvd;
  } SDR14REC;

typedef struct {  /*0x11 = FRU Locator*/
  ushort recid;
  uchar  sdrver;  /*usu. 0x51 = v1.5*/
  uchar  rectype; /* 01, 02, 11, 12, c0 */
  uchar  reclen;
  uchar  dev_access_adr;
  uchar  dev_slave_adr;
  uchar  access_lun;
  uchar  chan_num;
  uchar  reserved;
  uchar  dev_type; 
  uchar  dev_typemod; 
  uchar  entity_id; 
  uchar  entity_inst; 
  uchar  oem; 
  uchar  id_strlen;
  uchar  id_string[16];
  } SDR11REC;

typedef struct {  /*0x12 = IPMB Locator*/
  ushort recid;
  uchar  sdrver;  /*usu. 0x51 = v1.5*/
  uchar  rectype; /* 01, 02, 11, 12, c0 */
  uchar  reclen;
  uchar  dev_slave_adr;
  uchar  chan_num;
  uchar  power_state;
  uchar  dev_capab; 
  uchar  reserved[3]; 
  uchar  entity_id; 
  uchar  entity_inst; 
  uchar  oem; 
  uchar  id_strlen;
  uchar  id_string[16];
  } SDR12REC;

typedef struct {  /*0xc0 = OEM Record*/
  ushort recid;
  uchar  sdrver;  /*usu. 0x51 = v1.5*/
  uchar  rectype; /* 01, 02, 11, 12, c0 */
  uchar  reclen;
  uchar  manuf_id[3]; /*Intel = 0x57,0x01,0x00 = 343.*/
  uchar  oem_data[60]; /* (reclen-3 bytes)*/
  } SDRc0REC;

uchar htoi(uchar *inhex)
{
   // char rghex[16] = "0123456789ABCDEF";
   uchar val;
   uchar c;
   c = inhex[0] & 0x5f;  /* force cap */
   if (c > '9') c += 9;  /* c >= 'A' */
   val = (c & 0x0f) << 4;
   c = inhex[1] & 0x5f;  /* force cap */
   if (c > '9') c += 9;  /* c >= 'A' */
   val += (c & 0x0f);
   return(val);
}

int 
GetSDRRepositoryInfo(void)
{
	uchar resp[MAX_BUFFER_SIZE];
	int sresp = MAX_BUFFER_SIZE;
	int rc;
	int nSDR;
	int freespace;
	uchar cc = 0;
	int   i;

	rc = ipmi_cmd(GET_SDR_REPINFO, NULL, 0, resp,&sresp, &cc, 0);
	if (fdebug) printf("ipmi_cmd status = %d\n",rc);
	if (rc != ACCESS_OK) return(rc);
	if (cc != 0) return(cc);

	nSDR = resp[1] + (resp[2] << 8);
	freespace = resp[3] + (resp[4] << 8);
	if (fdebug) {
	   //successful, show data
           printf("SDR Repository (len=%d): ",sresp);
           for (i = 0; i < sresp; i++) printf("%02x ",resp[i]);
           printf("\n");
	   printf("SDR Info: nSDRs = %d, free space = %x\n",nSDR,freespace);
	}
	if (nSDR == 0) {
	   printf("SDR Repository is empty\n");
	   exit(0);
	}

	return(0);
}  /*end GetSDRRepositoryInfo()*/

int 
GetSensorThresholds(int sens_num, uchar *thr_data)
{
	uchar resp[MAX_BUFFER_SIZE];
	int sresp = MAX_BUFFER_SIZE;
	uchar inputData[6];
	int rc;
	uchar cc = 0;

	inputData[0] = sens_num;
	rc = ipmi_cmd(GET_SENSOR_THRESHOLD, inputData, 1, resp,&sresp, &cc, 0);
	if (fdebug)
		printf("GetSensorThreshold[%02x] rc = %d, resp(%d) %02x %02x %02x %02x %02x %02x %02x\n",
			sens_num,rc, sresp,resp[0],resp[1],resp[2],resp[3],
			resp[4],resp[5],resp[6]);
	if (rc != ACCESS_OK) return(rc);
	if (cc != 0) return(cc);
	if (sresp == 0) return(-2);
	memcpy(thr_data,resp,sresp);
	return(0);
}

int 
SetSensorThresholds(uchar sens_num, uchar hi, uchar lo, uchar *thr_data)
{
	uchar resp[MAX_BUFFER_SIZE];
	int sresp = MAX_BUFFER_SIZE;
	uchar inputData[8];
	int rc;
	uchar cc = 0;
	uchar sets = 0;
	int i;

	/* 
	 * Set the sensor Hysteresis before setting the threshold.
	 */
	memset(inputData,0,8);
	inputData[0] = sens_num;
	inputData[1] = 0xff;
	rc = ipmi_cmd(GET_SENSOR_HYSTERESIS, inputData, 2, resp,&sresp, &cc, 0);
	if (fdebug)
		printf("GetSensorHysteresis(%02x) rc = %d, cc = %x %02x %02x\n",
			sens_num,rc,cc,resp[0],resp[1]);
	if (rc != ACCESS_OK) return(rc);
	inputData[0] = sens_num;
	inputData[1] = 0xff;
	inputData[2] = resp[0];
	inputData[3] = resp[1];
	rc = ipmi_cmd(SET_SENSOR_HYSTERESIS, inputData, 4, resp,&sresp, &cc, 0);
	if (fdebug)
		printf("SetSensorHysteresis(%02x) rc = %d, cc = %x\n",
			sens_num,rc,cc);
	if (rc != ACCESS_OK) return(rc);
	
	/* 
	 * The application should validate that values are ordered, 
	 * e.g.  upper critical should be greater than upper 
	 * non-critical.
	 * Due to the limited command line parameter interface, 
	 * use the hi & lo values to set each of the thresholds.
	 * For a full implemenation, these thresholds should be set
	 * individually.
	 */
	memset(inputData,0,8);
	inputData[0] = sens_num;
	sets = thr_data[0];
	if (lo == 0xff) sets &= 0x38;  /* don't set lowers */
	else {
	   inputData[2] = lo + 2;  /* lower non-crit */
	   inputData[3] = lo + 1;  /* lower critical */
	   inputData[4] = lo;      /* lower non-recov */
	}
	if (hi == 0xff) sets &= 0x07;  /* don't set uppers */
	else {
	   inputData[5] = hi;      /* upper non-crit */
	   inputData[6] = hi + 1;  /* upper critical */
	   inputData[7] = hi + 2;  /* upper non-recov */
	}
	inputData[1] = sets;  /* which ones to set */
	{   /* show from/to changes */
		printf("GetThreshold[%02x]: %02x ",sens_num,sens_num);
		for (i = 0; i < 7; i++) printf("%02x ",thr_data[i]);
		printf("\n");
		printf("SetThreshold[%02x]: ",sens_num);
		for (i = 0; i < 8; i++) printf("%02x ",inputData[i]);
		printf("\n");
	}
	rc = ipmi_cmd(SET_SENSOR_THRESHOLD, inputData, 8, resp,&sresp, &cc, 0);
	if (fdebug)
		printf("SetSensorThreshold(%02x) rc = %d, cc = %x\n",
			sens_num,rc,cc);
	if (rc != ACCESS_OK) return(rc);
	if (cc != 0) return(cc);
	return(0);
}

int 
GetSensorReading(int sens_num, uchar *sens_data)
{
	uchar resp[MAX_BUFFER_SIZE];
	int sresp = MAX_BUFFER_SIZE;
	uchar inputData[6];
	int rc;
	uchar cc = 0;

	inputData[0] = sens_num;
	rc = ipmi_cmd(GET_SENSOR_READING, inputData, 1, resp,&sresp, &cc, 0);
	if (fdebug) 
		printf("ipmi_cmd %x status = %d, resp %02x %02x %02x %02x\n",
			GET_SENSOR_READING,rc,resp[0],resp[1],resp[2],resp[3]);
	if (rc != ACCESS_OK) return(rc);
	if (cc != 0) return(cc);

	/* only returns 4 bytes, no matter what type */
	memcpy(sens_data,resp,4);
	return(0);
}  /*end GetSensorReading()*/

int 
GetSensorReadingFactors(int snum, int raw, int *m, int *b, int *  b_exp, 
			int *r, int *a)
{
	uchar resp[MAX_BUFFER_SIZE];
	int sresp = MAX_BUFFER_SIZE;
	uchar inputData[6];
	int rc;
	uchar cc = 0;
	int  toler, a_exp;

	inputData[0] = snum;
	inputData[1] = raw;
	rc = ipmi_cmd(GET_SENSOR_READING_FACTORS, inputData, 2, 
                      resp,&sresp, &cc, 0);
	if (fdebug) printf("ipmi_cmd status = %d\n",rc);
	if (rc != ACCESS_OK) return(rc);
	if (cc != 0) return(cc);

        /* successful, copy values */
	*m = resp[1] + ((resp[2] & 0xc0) << 2);
	toler = resp[2] & 0x3f;
	*b = resp[3] + ((resp[4] & 0xc0) << 2);
	*a = (resp[4] & 0x3f) + ((resp[5] & 0xf0) << 4);
	a_exp = (resp[5] & 0xc0) >> 2;
	*r = (resp[6] &0xf0) >> 4;
	*b_exp = resp[6] & 0x0f;
	if (fdebug) {
	      printf("factors: next=%x m=%d b=%d b_exp=%d a=%d a_exp=%d r=%d\n",
			resp[0],*m,*b,*b_exp,*a,a_exp,*r);
	}
	return(0);
}

int 
GetSensorType(int snum, uchar *stype, uchar *rtype)
{
	uchar resp[MAX_BUFFER_SIZE];
	int sresp = MAX_BUFFER_SIZE;
	uchar inputData[6];
	int rc;
	uchar cc = 0;

	inputData[0] = snum;
	rc = ipmi_cmd(GET_SENSOR_TYPE, inputData, 1, 
                      resp,&sresp, &cc, 0);
	if (fdebug) printf("ipmi_cmd status = %d\n",rc);
	if (rc != ACCESS_OK) return(rc);
	if (cc != 0) return(cc);
        /* successful, copy values */
	*stype = resp[0];
	*rtype = resp[1] & 0x7f;
	return(rc);
}

int 
GetSDR(int recid, int *recnext, uchar *recdata,int srecdata)
{
	int sresp = MAX_BUFFER_SIZE;
	uchar resp[MAX_BUFFER_SIZE];
	uchar respchunk[18];
	uchar resid[2] = {0,0};
	uchar inputData[6];
	uchar cc = 0;
	int rc = -1;
	int  i, chunksz, thislen, off;
  
	chunksz = 16;
	// chunksz = 8;
	off = 0;
	if (srecdata > chunksz) {
	   sresp = MAX_BUFFER_SIZE;
	   rc = ipmi_cmd(RESERVE_SDR_REP, NULL, 0, resp, &sresp, &cc, 0);
	   resid [0] = resp [0]; resid [1] = resp [1];
	   if (fdebug) printf("ipmi_cmd status = %d, cc = %x\n",rc,cc);
	}
	for (off = 0; off < srecdata; )
	{
	   thislen = chunksz;
	   if ((off+chunksz) > srecdata) thislen = srecdata - off;
	   inputData[0] = resid[0];     /*res id LSB*/
	   inputData[1] = resid[1];     /*res id MSB*/
	   inputData[2] = recid & 0xff;  		/*record id LSB*/
	   inputData[3] = (recid & 0xff00) >> 8;	/*record id MSB*/
	   inputData[4] = off;      /*offset */
	   inputData[5] = thislen;  /*bytes to read, ff=all*/
	   sresp = thislen+10;
	   if (fdebug) printf("ipmi_cmd SDR id=%d off=%d, len=%d\n",
				recid,off,thislen);
	   rc = ipmi_cmd(GET_SDR, inputData, 6, respchunk, &sresp,&cc, fdebug);
	   if (fdebug) printf("ipmi_cmd SDR data[%d] status = %d, cc = %x\n",
			off,rc,cc);
	   if (rc != ACCESS_OK) return(rc);
	   if (cc != 0) return(cc);
	   /* successful */
	   memcpy(&resp[off],&respchunk[2],thislen);
	   if (off == 0) 
		*recnext = respchunk[0] + (respchunk[1] << 8);
	   off += thislen;
	}
	if (fdebug) {
	   printf("GetSDR (len=%d): ",srecdata);
	   for (i = 0; i < srecdata; i++) printf("%02x ",resp[2+i]);
	   printf("\n");
	   }
	memcpy(recdata,&resp[0],srecdata);
	return(rc);
}

int
bitnum(uchar value)
{
   int ret = 0;
   int i;
   /* returns the highest bit number number set in this byte. */
   for (i = 0; i < 8; i++) {
	if (value & 0x01) ret = i+1;  /*was ret++;*/
	value = (value >> 1);
   }
   return(ret);
}

static double
expon(int x, char y)
{
   double res;
   int i;
   /* compute exponent: x to the power y */
   res = 1;
   if (y > 0) {
	for (i = 0; i < y; i++) res = res * x;
   } else if (y < 0) {
	for (i = 0; i > y; i--) res = res / x;
   } /* else if if (y == 0) do nothing */
   return(res);
}

double
decodeValue(uchar raw, uchar *psdr)
{
   double floatval;
   int m, b, a, ax; 
   char r, b_exp;
   SDR01REC *sdr;
   char signval;

   sdr = (SDR01REC *)psdr;
   floatval = (double)raw;  /*default*/

   if (sdr->rectype == 0x01) {  /* SDR rectype == full */
	if (fdebug)
	   printf("units=%x base=%d mod=%d (raw=%d, nom_rd=%d)\n",
		sdr->sens_units,sdr->sens_base,sdr->sens_mod,
		raw, sdr->nom_reading);
	m = sdr->m + ((sdr->m_t & 0xc0) << 2);
	b = sdr->b + ((sdr->b_a & 0xc0) << 2);
	if (b & 0x0200) b |= 0xFFFFFC00;
	r = (sdr->rx_bx & 0xf0) >> 4;
	if (r & 0x08) r += 0xf0;  /* negative */
	a = (sdr->b_a & 0x3f) + ((sdr->a_ax & 0xf0) << 2);
	ax = (sdr->a_ax & 0x0c) >> 2;
	b_exp = sdr->rx_bx & 0x0f;

	if ((sdr->sens_units & 0xc0) == 0) {  /*unsigned*/
		floatval = (double)raw;
	} else {  /*signed*/
		signval = (char)raw;
		floatval = (double)signval;
	}
	floatval *= (double) m;
#ifdef MATH_OK
	floatval += (b * pow (10,b_exp));
	floatval *= pow (10,r);
#else
	floatval += (b * expon (10,b_exp));
	floatval *= expon (10,r);
#endif
	if (fdebug)
	   printf("decode1: m=%d b=%d b_exp=%x r=%d, a=%d ax=%d l=%x, floatval=%f\n",
			m,b,b_exp,r,a,ax,sdr->linear,floatval);
	if (raw != 0) {   /* avoid dividing by zero */
	  if (sdr->linear == 7) floatval = 1 / floatval;
	  /* else method not implemented, pass through */
	}
   }

#ifdef NOT_LINEAR
   {
     double be, re;
     rc = GetSensorType(sdr->sens_num,&stype,&rtype);
     if (fdebug)
	printf("decode: rc=%x, stype=%x, rtype=%x\n",rc,stype,rtype);
     if (rc != 0) return(floatval);

     /* GetSensorReadingFactors */
     rc = GetSensorReadingFactors(sdr->sens_num,raw,&m,&b,&b_exp,&r,&a);
     if (rc == 0) {
	// floatval = ((m * raw) + (b * be)) * re;
     } 
     if (fdebug) printf("decode: rc=%x, floatval=%f\n",rc,floatval);
   }
#endif

   return(floatval);
}

char *
decode_itype(uchar itype)
{
   char *retstr;
   int i;
   /* Decode the Interrupt Type from Entity Assoc records */

   retstr = tmpstr;
   if (itype <= 0x0f) sprintf(retstr,"IRQ_%d",itype);
   else if (itype <= 0x13) {
	strcpy(retstr,"PCI-A");
	for (i=0x10;i<itype;i++) retstr[4]++;
	}
   else if (itype == 0x14) strcpy(retstr,"SMI");
   else if (itype == 0x15) strcpy(retstr,"SCI");
   else if (itype >= 0x20 && itype <= 0x5f) 
	sprintf(retstr,"SysInt_%d",itype-0x20);
   else if (itype == 0x60) strcpy(retstr,"ACPI/PnP");
   else if (itype == 0xFF) strcpy(retstr,"NoInt");
   else strcpy(retstr,"Invalid");
   return(retstr);
}

void
ShowThresh(int flg, uchar bits, uchar *vals, uchar *sdr)
{
   char str[120] = "";
   char part[20]; /* ~12 bytes used */
   double ival;
   if (flg) {  /* Compact, did GetThresholds, reverse order */
	if (bits & 0x20) {
		ival = decodeValue(vals[5],sdr);
		sprintf(part,"hi-norec %.2f ",ival);
		strcat(str,part);
	}
	if (bits & 0x10) {
		ival = decodeValue(vals[4],sdr);
		sprintf(part,"hi-crit %.2f ", ival);
		strcat(str,part);
	}
	if (bits & 0x08) {
		ival = decodeValue(vals[3],sdr);
		sprintf(part,"hi-noncr %.2f ",ival);
		strcat(str,part);
	}
	if (bits & 0x04) {
		ival = decodeValue(vals[2],sdr);
		sprintf(part,"lo-norec %.2f ",ival);
		strcat(str,part);
	}
	if (bits & 0x02) {
		ival = decodeValue(vals[1],sdr);
		sprintf(part,"lo-crit %.2f ", ival);
		strcat(str,part);
	}
	if (bits & 0x01) {
		ival = decodeValue(vals[0],sdr);
		sprintf(part,"lo-noncr %.2f ",ival);
		strcat(str,part);
	}
   } else {  /*Full SDR*/
	if (fdebug) printf("ShowThresh[%x]: bits = %02x\n",sdr[7],bits);
	if (bits & 0x20) {
		ival = decodeValue(vals[0],sdr);
		sprintf(part,"hi-norec %.2f ",ival);
		strcat(str,part);
	}
	if (bits & 0x10) {
		ival = decodeValue(vals[1],sdr);
		sprintf(part,"hi-crit %.2f ",ival);
		strcat(str,part);
	}
	if (bits & 0x08) {
		ival = decodeValue(vals[2],sdr);
		sprintf(part,"hi-noncr %.2f ",ival);
		strcat(str,part);
	}
	if (bits & 0x04) {
		ival = decodeValue(vals[3],sdr);
		sprintf(part,"lo-norec %.2f ",ival);
		strcat(str,part);
	}
	if (bits & 0x02) {
		ival = decodeValue(vals[4],sdr);
		sprintf(part,"lo-crit  %.2f ",ival);
		strcat(str,part);
	}
	if (bits & 0x01) {
		ival = decodeValue(vals[5],sdr);
		sprintf(part,"lo-noncr %.2f ",ival);
		strcat(str,part);
	}
   }
   printf("\t%s\n",str);
}

void
ShowSDR(uchar *sdr)
{
  SDR01REC *sdr01;
  SDR02REC *sdr02;
  SDR08REC *sdr08;
  SDR11REC *sdr11;
  SDR12REC *sdr12;
  SDR14REC *sdr14;
  SDRc0REC *sdrc0;
  char idstr[17];
  long mfgid;
  int len, ilen, i, j;
  int ioff;
  uchar sens[4];
  uchar c, sens_cap;
  int rc; 
  double val;

  len = sdr[4] + 5;
  sens_cap = 0x80;  /*ignore*/
  if (fdebug) printf("ShowSDR: len=%d, type=%x\n",len,sdr[3]);
  memset(sens,0,4);
  switch(sdr[3])
  {
    case 0x01:   /* Full sensor record */
	sdr01 = (SDR01REC *)sdr;
	ioff = 48;
	if (ioff > len) {
		if (fdebug) printf("bad length: type=%x, len=%d, ioff=%d\n",
					sdr[3],len,ioff);
		printf("Please apply the correct FRU/SDR diskette\n");
		return;
	}
	sens_cap = sdr[12];
	ilen = len - ioff;
	memcpy(idstr,&sdr[ioff],ilen);
	for (i=ilen; i<16; i++) idstr[i] = ' ';
	idstr[16] = 0;  /* stringify */
	if (fdebug) printf("ilen=%d, istr0=%c, sizeof=%d, s0=%x\n",
				ilen,idstr[0],sizeof(SDR01REC),sdr[ioff]);
	rc = GetSensorReading(sdr01->sens_num,sens);
	/* if rc != 0, leave sens values zero */
	i = bitnum(sens[2]);
	if (fdebug) printf("bitnum = %d\n",i);
	val = decodeValue(sens[0],sdr);
	if (sdr01->sens_base <= NUNITS) j = sdr01->sens_base;
	else j = 0;
	if (flist) {
		printf("Full sensor[%04x]: num %02x %s = %s %.2f %s\n",
			sdr01->recid, sdr01->sens_num,idstr,
			sensor_dstatus[i],val,unit_types[j]);
	} else
	printf("SDR Full %04x %02x %02x sensor: %02x %02x %s = %02x %s %.2f %s\n",
		sdr01->recid, sdr01->rectype, sdr01->reclen,
		sdr01->sens_ownid, sdr01->sens_num, idstr,
		sens[0], sensor_dstatus[i], 
		val, unit_types[j]);
	if (fdebug && fshowthr)
		   printf("cap=%02x settable=%02x, readable=%02x\n",
			  sens_cap,sdr[19],sdr[20]);
	if (fshowthr && (sens_cap & 0x0f) != 0x03) {
		/* Thresholds, so show them */
		/* Use settable bits to show thresholds, since the 
		   readable values will be off for Full SDRs.  */
		ShowThresh(0,sdr[19],&sdr[36],sdr); 
		// ShowThresh(0,0x3f,&sdr[36],sdr); /* to show all %%%% */
	}
	break;
    case 0x02:   /* Compact sensor record */
	sdr02 = (SDR02REC *)sdr;
	ioff = 32;
	if (ioff > len) {
		if (fdebug) printf("bad length: type=%x, len=%d, ioff=%d\n",
					sdr[3],len,ioff);
		printf("Please apply the correct FRU/SDR diskette\n");
		return;
	}
	sens_cap = sdr[12];
	ilen = len - ioff;
	memcpy(idstr,&sdr[ioff],ilen);
	for (i=ilen; i<16; i++) idstr[i] = ' ';
	idstr[16] = 0;  /* stringify */
	if (fdebug) printf("ilen=%d, istr0=%c, sizeof=%d, s0=%x\n",
				ilen,idstr[0],sizeof(SDR02REC),sdr[ioff]);
	memset(sens,0,sizeof(sens));
	rc = GetSensorReading(sdr02->sens_num,sens);
	/* if rc != 0, leave sens values zero */
	if (sdr02->sens_ownid == 0xc0) { /*HSC*/
		if (sdr02->sens_num == 0x02 ||
		    sdr02->sens_num == 0x03 ) {  /*dev status*/
		   if ((sens[2] & 0x20) == 0)  i = 11; /*ready*/
		   else i = 12; /*faulty*/
		} else {  /* 09 & 0a = dev presence*/
		   if ((sens[2] & 0x20) == 0) i = 9; /*present*/
		   else i = 10; /*missing*/
		}
	} else   /*BMC*/
		i = bitnum(sens[2]);
	if (fdebug) printf("sens_num %x reading rc = %d, bitnum = %d\n",
				sdr02->sens_num,rc,i);
	if (flist)
		printf("Compact sensor[%04x]: num %02x %s = %s\n",
			sdr02->recid, sdr02->sens_num,idstr,sensor_dstatus[i]);
	else
	printf("SDR Comp %04x %02x %02x sensor: %02x %02x %s = %02x %02x %02x %02x %s\n",
		sdr02->recid, sdr02->rectype, sdr02->reclen, 
		sdr02->sens_ownid, sdr02->sens_num, idstr,
		sens[0], sens[1], sens[2], sens[3], sensor_dstatus[i]);
	if (fdebug && fshowthr)
		   printf("cap=%02x settable=%02x, readable=%02x\n",
			  sens_cap,sdr[19],sdr[20]);
	if (fshowthr &&
	    ((sens_cap & 0x80) == 0) && (sens_cap & 0x0C) != 0) {
		uchar thresh[7];
		/* Thresholds, show them */
		/* Use readable bits to get & show thresholds */
		if (sdr[20] != 0) {
		   rc = GetSensorThresholds(sdr02->sens_num,&thresh[0]);
		   if (rc == 0) ShowThresh(1,thresh[0],&thresh[1],sdr);
		}
	}
	break;
    case 0x08:   /* Entity Association record */
	sdr08 = (SDR08REC *)sdr;
	if (!flist)
	{
	printf("SDR Asso %04x %02x %02x %02x %02x %02x: ",
		sdr08->recid, sdr08->rectype, sdr08->reclen, 
		sdr08->contid, sdr08->continst, sdr08->flags);
	for (i = 0; i < 8; i++) printf("%02x ",sdr08->edata[i]);
	printf("\n");
	}
	break;
    case 0x11:   /* FRU record */
	sdr11 = (SDR11REC *)sdr;
	ioff = 16;
	if (ioff > len) {
		if (fdebug) printf("bad length: type=%x, len=%d, ioff=%d\n",
					sdr[3],len,ioff);
		printf("Please apply the correct FRU/SDR diskette\n");
		return;
	}
	ilen = len - ioff;
	memcpy(idstr,&sdr[ioff],ilen);
	idstr[ilen] = 0;  /* stringify */
	if (fdebug) printf("ilen=%d, istr0=%c, sizeof=%d, s0=%x\n",
				ilen,idstr[0],sizeof(SDR11REC),sdr[ioff]);
	if (flist)
		printf("FRU record[%x]: device %02x %s\n",
			sdr11->recid, sdr11->dev_slave_adr,idstr);
	else
	printf("SDR FRU  %04x %02x %02x dev: %02x %02x %02x %02x %s\n",
		sdr11->recid, sdr11->rectype, sdr11->reclen,
		sdr11->dev_slave_adr, sdr11->chan_num,
		sdr11->entity_id, sdr11->entity_inst, 
		idstr);
	break;
    case 0x12:   /* IPMB record */
	sdr12 = (SDR12REC *)sdr;
	ioff = 16;
	if (ioff > len) {
		if (fdebug) printf("bad length: type=%x, len=%d, ioff=%d\n",
					sdr[3],len,ioff);
		printf("Please apply the correct FRU/SDR diskette\n");
		return;
	}
	ilen = len - ioff;
	memcpy(idstr,&sdr[ioff],ilen);
	idstr[ilen] = 0;  /* stringify */
	if (fdebug) printf("ilen=%d, istr0=%c, sizeof=%d, s0=%x\n",
				ilen,idstr[0],sizeof(SDR12REC),sdr[ioff]);
	if (flist)
		printf("IPMB record[%x]: addr %02x %02x %s\n",
			sdr12->recid, sdr12->dev_slave_adr, 
			sdr12->chan_num,idstr);
	else
	printf("SDR IPMB %04x %02x %02x dev: %02x %02x %02x %02x %s\n",
		sdr12->recid, sdr12->rectype, sdr12->reclen,
		sdr12->dev_slave_adr, sdr12->chan_num,
		sdr12->entity_id, sdr12->entity_inst, 
		idstr);
	break;
    case 0x14:   /* BMC Message Channel Info record */
	sdr14 = (SDR14REC *)sdr;
	if(!flist){
	printf("SDR BMsg %04x %02x %02x: ",
		sdr14->recid, sdr14->rectype, sdr14->reclen );
	for (i = 0; i < 8; i++) printf("%02x ",sdr14->mdata[i]);
	printf("%s %s %02x\n",decode_itype(sdr14->mint),
		decode_itype(sdr14->eint), sdr14->rsvd);
	}
	break;
    case 0xc0:   /* OEM record (#343 = Intel) */
	sdrc0 = (SDRc0REC *)sdr;
	if(!flist)
	{
	mfgid = sdrc0->manuf_id[0] + (sdrc0->manuf_id[1] << 8) 
		+ (sdrc0->manuf_id[2] << 16);
	printf("SDR OEM  %04x %02x %02x manuf=", 
		sdrc0->recid, sdrc0->rectype, sdrc0->reclen);
	if (mfgid == 343) printf("Intel: ");
	else printf("%ld: ",mfgid);
	c = sdr[8];  /*OEM subtype*/
        if (c == 0x53) {   /* SDR version subtype (has ASCII) */
	    for (i = 8; i < len; i++) {
		c = sdr[i];
		if (c < 0x20 || c > 0x7f) printf("[%02x]",c);
		else printf("%c",c);
	    }
	} else {  /* subtypes 02,06,07,60 */
	    for (i = 8; i < len; i++) 
		printf("%02x ",sdr[i]);
	}
	printf("\n");
	}
	break;
    default:
	/* also saw type = 0x08 & 0x14 on STL2s */
	if (!flist){
	printf("SDR (type=0x%02x): ", sdr[3]);
	for (i = 0; i < len; i++) printf("%02x ",sdr[i]);
	printf("\n");
  }
  }
  return;
}

int
GetPowerOnHours(void)
{
	uchar resp[MAX_BUFFER_SIZE];
	int sresp = MAX_BUFFER_SIZE;
	uchar cc;
	int rc = -1;
	int i;
  
	sresp = MAX_BUFFER_SIZE;
	rc = ipmi_cmd(GET_POWERON_HOURS, NULL, 0, resp, &sresp, &cc, 0);
	if (rc == 0) {
	   unsigned long *hrs;
	   /* show the hours */
	   hrs = (unsigned long *)&resp[1];
           printf("SDR IPMI \t    sensor: Power On Hours \t   = %ld hours\n",
			*hrs);
	}
	if (fdebug) {
           printf("PowerOnHours (rc=%d len=%d): ",rc,sresp);
	   if (rc == 0) 
           for (i = 0; i < sresp; i++) printf("%02x ",resp[i]);
           printf("\n");
	}
	return(rc);
}

#ifdef NOT
/* not valid yet */
#define NRSSA   20
uchar salist[NRSSA];
int nrssa = NRSSA;
int nret = 0;

int Discovery(uchar *list, int max, int *numret)
{
   uchar devrec[16];
   int i, ret, sresp;
   uchar sa, cc;
   uchar ipmi_maj, ipmi_min;

   i = 0;
   if (fdebug) printf("Discovery: ");
   for (sa = 0x20; sa < 0xF0; sa+=4) {
	sresp = sizeof(devrec);
	ret = ipmi_cmd_raw(GET_DEVICE_ID,NETFN_STOR,sa, 0/*bus*/, 0/*BMC_LUN*/,
		   NULL,0,&devrec[0], &sresp,&cc,fdebug);
	if (ret == 0 && cc == 0) {
	   ipmi_maj = devrec[4] & 0x0f;
	   ipmi_min = devrec[4] >> 4;
	   if (fdebug) printf("%02x ",sa);
	   salist[i++] = sa;
	}
   }
   if (fdebug) printf("\n     numsa = %d",i);
   *numret = i;
   if (i > 0) return(0);
   else return(-1);
}
#endif

int
main(int argc, char **argv)
{
   int ret, numsa;
   char c;
   int recid, recnext;
   uchar sdrdata[MAX_BUFFER_SIZE];
   uchar devrec[16];
   int sz, i;

   printf("%s: version %s\n",progname,progver);

   while ( (c = getopt( argc, argv,"stxh:l:n:?")) != EOF )
      switch(c) {
	  case 's': flist = 1;	break;
	  case 't': fshowthr = 1; break;
	  case 'n': 
		if (strncmp(optarg,"0x",2) == 0) i = htoi(&optarg[2]);
		else i = atoi(optarg);
		sensor_num = (uchar)i;
		break;
	  case 'h': 
		if (strncmp(optarg,"0x",2) == 0) i = htoi(&optarg[2]);
		else i = atoi(optarg);
		sensor_hi = (uchar)i;
		fsetthresh = 1;
		break;
	  case 'l': 
		if (strncmp(optarg,"0x",2) == 0) i = htoi(&optarg[2]);
		else i = atoi(optarg);
		sensor_lo = (uchar)i;
		fsetthresh = 1;
		break;
          case 'x': fdebug = 1;   break;
          default:
                printf("Usage: %s [-h -l -n -s -t -x]\n", progname);
                printf("where -x      shows eXtra debug messages\n");
                printf("      -s      displays a Simpler output format\n");
                printf("      -t      shows Threshold values also\n");
                printf("      -n num  specifies the sensor Number to set\n");
                printf("      -h num  specifies the High threshold to set\n");
                printf("      -l num  specifies the Low threshold to set\n");
                exit(1);
      }               

   ret = ipmi_getdeviceid(devrec,sizeof(devrec));
   if (ret == 0) {
	uchar ipmi_maj;
	uchar ipmi_min;
	ipmi_maj = devrec[4] & 0x0f;
	ipmi_min = devrec[4] >> 4;
	printf("-- BMC version %x.%x, IPMI version %d.%d \n",
		devrec[2],  devrec[3], ipmi_maj, ipmi_min);
   } else {
	printf("ipmi_getdeviceid error, ret = %d\n",ret);
	exit(1);
   }

   ret = GetSDRRepositoryInfo();
   if (fdebug) printf("GetSDRRepositoryInfo: ret = %x\n",ret);

   recid = 0;
   while (recid != 0xffff) {
	/* Get partial record */
   	sz = 16;
   	// sz = 8;
	if (fdebug) printf("GetSDR: part recid = %x, sz = %d\n",recid,sz);
	ret = GetSDR(recid, &recnext, sdrdata, sz);
	if (ret != 0) break;
	sz = sdrdata[4] + 5;     /*now get full record*/
	if (sz > sizeof(sdrdata)) sz = sizeof(sdrdata);
	if (fdebug) printf("GetSDR: full recid = %x, subtype=%02x max=%d sz = %d\n",
			recid,sdrdata[8],MAX_BUFFER_SIZE,sz);
	ret = GetSDR(recid, &recnext, sdrdata, sz);
	if (fdebug) printf("GetSDR: full ret = %x, next=%x\n",ret,recnext);
	if (ret != 0) break;
	if (fdebug) {
	   printf("SDR (sz=%d):\t",sz);
	   for (i = 0; i < sz; i++) {
		if (i > 0 && i % 20 == 0) printf("\n\t");
		printf("%02x ",sdrdata[i]);
	   }
	   printf("\n");
	   }
	ShowSDR(sdrdata);

	if (fsetthresh && (sdrdata[7] == sensor_num)) {
	   printf("\tSetting SDR %04x sensor %02x to lo=%02x hi=%02x\n",
		  recid,sensor_num,sensor_lo,sensor_hi);
	}

	recid = recnext;
   }
   if (ret != 0) printf("SDR error = %x\n",ret);
   ret = GetPowerOnHours();

   if (fsetthresh) {
	uchar tdata[7];
	ret = GetSensorThresholds(sensor_num,tdata);
	if (ret != 0) exit(ret);
#ifdef TEST
	printf("thresh(%02x): %02x %02x %02x %02x %02x %02x %02x %02x\n",
		sensor_num, sensor_num, tdata[0], tdata[1], tdata[2],
		tdata[3], tdata[4], tdata[5], tdata[6]);
	printf("   set(%02x):          %02x       %02x \n",
		sensor_num,sensor_lo,sensor_hi);
#endif
	ret = SetSensorThresholds(sensor_num,sensor_hi,sensor_lo,tdata);
	printf("SetSensorThreshold[%02x] to lo=%02x hi=%02x, ret = %d\n",
		sensor_num,sensor_lo,sensor_hi,ret);
   }
   exit(ret);
}

/* end sensor.c */
