/*
 * showsel.c
 *
 * This tool reads the firmware System Event Log records via IPMI commands.
 * It can use either the Intel /dev/imb driver or VALinux /dev/ipmikcs.
 *
 * Author:  Andy Cress  andrew.r.cress@intel.com
 *
 * Copyright (c) 2001-2002 Intel Corporation. 
 *
 * 10/16/01 Andy Cress - created
 * 10/19/01 Andy Cress - added text for Sensor Types from Table 36-3
 * 10/24/01 Andy Cress - added panic string display for OS Crit Stop
 * 11/01/01 Andy Cress - added logic to write SEL records to syslog
 * 11/15/01 Andy Cress - added logic for more description strings
 * 01/31/02 Andy Cress - isolated dependencies on /dev/imb to ipmi_cmd_ia(),
 *                       added ipmi_cmd_va() for va ipmi driver also,
 *                       added posterrs array for more descriptions.
 * 02/06/02 Andy Cress - fixed bug 279 (extra debug msg at EOF).
 * 03/25/02 Andy Cress - show free space, add -c to clear SEL.
 * 04/11/02 Andy Cress - decode timestamp into readable form
 * 06/14/02 Andy Cress - also show status error in ClearSEL
 * 07/02/02 Andy Cress v1.3 add more Usage text
 * 08/02/02 Andy Cress v1.4 moved common ipmi_cmd() code to ipmicmd.c 
 * 10/07/02 Andy Cress v1.5 added -v option with BMC version too
 * 01/16/03 Andy Cress v1.6 Handle new OS crit stop format with die code
 * 01/29/03 Andy Cress v1.7 added MV OpenIPMI support
 * 02/05/03 Andy Cress v1.8 show better message if empty SEL (cc=0xCB),
 *                          show an additional warning if free space is low.
 * 02/18/03 Andy Cress v1.9 trim out some fields so it fits on 1 line,
 *                          decode Boot Event subcodes
 * 02/27/03 Andy Cress v1.10 change OS Crit Stop decoding to handle new types
 * 03/20/03 Andy Cress v1.11 for -w, save id also, so we don't have to start
 *                           over at the beginning each time.
 * 04/30/03 Andy Cress v1.12 changed display ordering.
 * 06/23/03 Andy Cress v1.13 fix -w if log gets cleared
 * 08/19/03 Andy Cress v1.14 handle OEM/other record types
 * 09/16/03 Andy Cress v1.15 added more sens_desc strings
 * 10/03/03 Andy Cress v1.16 added more sens_desc strings (for boot)
 * 01/19/04 Andy Cress v1.17 added more sens_desc for Fans
 */
/*M*
Copyright (c) 2002, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#include <string.h>
#include <time.h>
#include <getopt.h>

#include "ipmicmd.h" 

#define  SELprintf  printf
 
typedef struct
{
        short              record_id;
        uchar              record_type;
        uint               timestamp;
        short              generator_id;
        uchar              evm_rev;         //event message revision
        uchar              sensor_type;
        uchar              sensor_number;
        uchar              event_trigger;
        uchar              event_data1;
        uchar              event_data2;
        uchar              event_data3;
}       SEL_RECORD;

#define RECORD_BASE     2     //base value to the SEL record in IMB resp data
#define RID_OFFSET      0     //byte offset to the record id
#define RTYPE_OFFSET    2     //byte offset to the record type
#define RTS_OFFSET      3     //byte offset to the record timestamp
#define RGID_OFFSET     7     //byte offset to the record generator id
#define REREV_OFFSET    9     //byte offset to the record event message rev
#define RSTYPE_OFFSET   10    //byte offset to the record sensor type
#define RSN_OFFSET      11    //byte offset to the record sensor number
#define RET_OFFSET      12    //byte offset to the record event trigger
#define RDATA_OFFSET    13    //byte offset to the record event data  
#define MIN_FREE       128    /* minimal bytes of free SEL space */

char progname[] = "showsel";
char progver[] = "1.17";
char idxfile[] = "/usr/share/panicsel/sel.idx";
char fdebug = 0;
char fall = 1;
char fwritesel = 0;
char fclearsel = 0;
char fonlyver = 0;
uint savtime = 0;
short savid = 0;

static const char *sensor_types[] = {
/* 00h */ "reserved",
/* 01h */ "Temperature",
/* 02h */ "Voltage",
/* 03h */ "Current",
/* 04h */ "Fan",
/* 05h */ "Platform Chassis Intrusion",
/* 06h */ "Platform Security Violation",
/* 07h */ "Processor",
/* 08h */ "Power Supply",
/* 09h */ "Power Unit",
/* 0Ah */ "Cooling Device",
/* 0Bh */ "FRU Sensor",
/* 0Ch */ "Memory",
/* 0Dh */ "Drive Slot",
/* 0Eh */ "POST Memory Resize",
/* 0Fh */ "System Firmware",
/* 10h */ "SEL Disabled",
/* 11h */ "Watchdog 1",
/* 12h */ "System Event",          /* offset 0,1,2 */
/* 13h */ "Critical Interrupt",    /* offset 0,1,2 */
/* 14h */ "Button",                /* offset 0,1,2 */
/* 15h */ "Board",
/* 16h */ "Microcontroller",
/* 17h */ "Add-in Card",
/* 18h */ "Chassis",
/* 19h */ "Chip Set",
/* 1Ah */ "Other FRU",
/* 1Bh */ "Cable / Interconnect",
/* 1Ch */ "Terminator",
/* 1Dh */ "System Boot Initiated",
/* 1Eh */ "Boot Error",
/* 1Fh */ "OS Boot",
/* 20h */ "OS Critical Stop",
/* 21h */ "Slot / Connector",
/* 22h */ "ACPI Power State",
/* 23h */ "Watchdog 2",
/* 24h */ "Platform Alert",
/* 25h */ "Entity Presence",
/* 26h */ "Monitor ASIC",
/* 27h */ "LAN",
/* 28h */ "Management Subsystem Health",
/* 29h */ "Battery",
};

static struct {
  unsigned char code0;
  unsigned char code1;
  char          *msg;
  } posterrs[] = {
   0x62, 0x01, "BIOS unable to apply BIOS update to processor 1"        ,
   0x63, 0x01, "BIOS unable to apply BIOS update to processor 2"        ,
   0x64, 0x01, "BIOS does not support current stepping for processor 1" ,
   0x65, 0x01, "BIOS does not support current stepping for processor 2" ,
   0x00, 0x02, "Failure Fixed Disk"                  ,
   0x10, 0x02, "Stuck Key"                           ,
   0x11, 0x02, "Keyboard error"                      ,
   0x12, 0x02, "Keyboard Controller Failed"          ,
   0x13, 0x02, "Keyboard locked"                     ,
   0x20, 0x02, "Monitor type does not match CMOS"    ,
   0x30, 0x02, "System RAM Failed at offset"         ,
   0x31, 0x02, "Shadow RAM Failed at offset"         ,
   0x32, 0x02, "Extended RAM Failed at offset"       ,
   0x50, 0x02, "System battery is dead"              ,
   0x51, 0x02, "System CMOS checksum bad"            ,
   0x60, 0x02, "System timer error"                  ,
   0x70, 0x02, "Real time clock error"               ,
   0x97, 0x02, "ECC Memory error in bank"            ,
   0xb2, 0x02, "Incorrect Drive A: type"             ,
   0xb3, 0x02, "Incorrect Drive B: type"             ,
   0xd0, 0x02, "System cache error - Cache disabled" ,
   0xf5, 0x02, "DMA Test Failed"                     ,
   0xf6, 0x02, "Software NMI Failed"                 ,
   0x01, 0x04, "Invalid System Configuration Data" ,
   0x03, 0x04, "Resource Conflict"                 ,
   0x04, 0x04, "Resource Conflict"                 ,
   0x05, 0x04, "Expansion ROM not initialized"     ,
   0x06, 0x04, "Warning: IRQ not configured"       ,
   0x04, 0x05, "Resource Conflict"             ,
   0x05, 0x05, "Expansion ROM not initialized" ,
   0x06, 0x05, "Warning: IRQ not configured"   ,
   0x01, 0x06, "Device configuration changed"          ,
   0x02, 0x06, "Configuration error - device disabled" ,
   0x00, 0x81, "processor 0 failed BIST"                            ,
   0x01, 0x81, "processor 1 failed BIST"                            ,
   0x04, 0x81, "processor 0 Internal Error (IERR) failure"          ,
   0x05, 0x81, "processor 1 Internal Error (IERR) failure"          ,
   0x06, 0x81, "processor 0 Thermal Trip failure"                   ,
   0x07, 0x81, "processor 1 Thermal Trip failure"                   ,
   0x08, 0x81, "Watchdog Timer failed on last boot, BSP switched"   ,
   0x0a, 0x81, "processor 1 failed initialization on last boot"     ,
   0x0b, 0x81, "processor 0 failed initialization on last boot"     ,
   0x0c, 0x81, "processor 0 disabled, system in Uni-processor mode" ,
   0x0d, 0x81, "processor 1 disabled, system in Uni-processor mode" ,
   0x0e, 0x81, "processor 0 failed FRB Level 3 timer"               ,
   0x0f, 0x81, "processor 1 failed FRB Level 3 timer"               ,
   0x10, 0x81, "Server Management Interface failed to function"     ,
   0x20, 0x81, "IOP sub-system is not functional"                   ,
   0x50, 0x81, "NVRAM (CMOS) Cleared by Jumper"                     ,
   0x51, 0x81, "NVRAM (CMOS) Checksum Error, NVRAM cleared"         ,
   0x52, 0x81, "NVRAM (CMOS) Data Invalid, NVRAM cleared"           ,
   0x00, 0x00, NULL
  };

#define NGDESC   6
struct {
 uchar g_id;
 const char desc[10];
} gen_desc[NGDESC] = {
  0x00, "IPMB",
  0x03, "BIOS",
  0x20, "BMC ",
  0x21, "SMI ",
  0x28, "CHAS",  /* Chassis Bridge Controller */
  0xc0, "HSC "   /* SAF-TE Hot-Swap Controller*/
};

#define NSDESC   19
struct {
 uchar s_typ;
 uchar s_num;
 uchar evtrg;
 uchar data1;
 uchar data2;
 uchar data3;
 const char desc[40];
} sens_desc[NSDESC] = {
 0x01, 0xff, 0xff, 0x57, 0xff, 0xff, "Upper Non-critical", /*temperature*/
 0x01, 0xff, 0xff, 0x59, 0xff, 0xff, "Upper Critical    ", /*temperature*/
 0x04, 0xff, 0x01, 0x50, 0xff, 0xff, "Lower Non-critical", /*fan*/
 0x04, 0xff, 0x01, 0x52, 0xff, 0xff, "Lower Critical    ", /*fan*/
		/*Note that last 2 bytes usu show actual & threshold values*/
 0x07, 0xff, 0xff, 0x41, 0xff, 0xff, "Thermal trip", /*processor*/
 0x09, 0x01, 0xff, 0xff, 0xff, 0xff, "Power Off/Down",
 0x0c, 0x08, 0xff, 0xff, 0xff, 0xff, "Correctable ECC",
 0x0f, 0x06, 0xff, 0xff, 0xff, 0xff, "POST Code",
 0x10, 0x09, 0xff, 0x42, 0x0f, 0xff, "Log Cleared",
 0x10, 0x09, 0xff, 0xc0, 0x03, 0xff, "ECC Memory Errors",
 0x12, 0x83, 0xff, 0x05, 0x00, 0xff, "Time/Clock Sync-1",
 0x12, 0x83, 0xff, 0x05, 0x80, 0xff, "Time/Clock Sync-2",
 0x12, 0x83, 0xff, 0x01, 0xff, 0xff, "OEM System Booted",
    /*  Often see these 3 Boot records with reboot:
     * 12 83 6f 05 00 ff = TimeStamp Clock Sync 1 (BIOS going down)
     * 12 83 6f 05 80 ff = TimeStamp Clock Sync 2 (BIOS is up)
     * 12 83 6f 01 ff ff = OEM System Boot Event  (Completed POST) */
 0x14, 0xff, 0xff, 0x42, 0xff, 0xff, "Reset Button pressed",
 0x14, 0xff, 0xff, 0x40, 0xff, 0xff, "Power Button pressed",
 0x23, 0x03, 0xff, 0xc1, 0xff, 0xff, "Hard Reset action",  /*watchdog2*/
 0x23, 0x03, 0xff, 0xc2, 0xff, 0xff, "Power down action",  /*watchdog2*/
 0xf3, 0x85, 0xff, 0x41, 0xff, 0xff, "State is now OK",
 0x20, 0x00, 0xff, 0xff, 0xff, 0xff, "OS Kernel Panic"
};
 

////////////////////////////////////////////////////////////////////////////////
//  GetSelEntry
////////////////////////////////////////////////////////////////////////////////
/*F*
//  Name			:	GetSelEntry
//
//  Purpose			:	This routine gets the next SEL record
//						
//  Parameters		:	
//	pRecordID		:	input/output pointer to next Record ID
//	selRecord		:	output pointer to the sel record
//
//  Returns			:	0 if success
//					-1 if failure.
//				(The error code will be in "g_iThreadLastError")
//						
//	Notes			:	uses ipmi_cmd()
//
*F*/
int GetSelEntry(short *pRecordID, SEL_RECORD *selRecord)
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar inputData[6];
	uchar completionCode;
	char* cpRecordID;	 //for next record id
	char* cpCurrentRID;  //for current record id
	char* cpCurrentTimeStamp; //for current record timestamp
	char* cpCurrentGID;  //for current record generator id
	short inRecordID;    //save the input record id

	inRecordID = *pRecordID;
	if (!selRecord) 
	{
		if (fdebug)
		    printf("GetSelEntry: error, output buffer is NULL\n");
		return -1;
	}

	//set the reservation id to zero
	inputData[0]            = 0;
	inputData[1]            = 0;

	//set the record id to get
	cpRecordID              = (char*) pRecordID;
	inputData[2]            = cpRecordID[0];
	inputData[3]            = cpRecordID[1];

	//set the offset to the record to value zero
	inputData[4]            = 0;

	//set the number of byte to read.  0xFF means read the entire record
	inputData[5]            = (char) 0xFF;

	status = ipmi_cmd(GET_SEL_ENTRY, inputData, 6, responseData, 
			&responseLength, &completionCode, fdebug);

	if (status == ACCESS_OK) {
		if( completionCode ) {
			if (completionCode == 0xCB && inRecordID == 0)
			  SELprintf("Firmware Log (SEL) is empty\n");
			else 
			  SELprintf("GetSelEntry: completion code=%x\n", 
				completionCode); // responseData[0]);
		} else {
			//successful, done
			//save the next SEL record id
			cpRecordID[0] = responseData[0];
			cpRecordID[1] = responseData[1];

			//save the SEL record content
			//save record id
			cpCurrentRID    = (char *) &(selRecord->record_id);
			cpCurrentRID[0] = responseData[RECORD_BASE+RID_OFFSET];
			cpCurrentRID[1] = responseData[RECORD_BASE+RID_OFFSET+1];

			//save record type
			selRecord->record_type = responseData[RECORD_BASE+RTYPE_OFFSET];

			//save record timestamp
			cpCurrentTimeStamp = (char *) &(selRecord->timestamp);
			cpCurrentTimeStamp[0] = responseData[RECORD_BASE+RTS_OFFSET];
			cpCurrentTimeStamp[1] = responseData[RECORD_BASE+RTS_OFFSET+1];
			cpCurrentTimeStamp[2] = responseData[RECORD_BASE+RTS_OFFSET+2];
			cpCurrentTimeStamp[3] = responseData[RECORD_BASE+RTS_OFFSET+3];

			//save record generator id
			cpCurrentGID    = (char *) &(selRecord->generator_id);
			cpCurrentGID[0] = responseData[RECORD_BASE + RGID_OFFSET];
			cpCurrentGID[1] = responseData[RECORD_BASE + RGID_OFFSET+1];

			//save record event message rev
			selRecord->evm_rev = responseData[RECORD_BASE + REREV_OFFSET];

			//save record sensor type
			selRecord->sensor_type = responseData[RECORD_BASE + RSTYPE_OFFSET];

			//save record sensor number
			selRecord->sensor_number= responseData[RECORD_BASE + RSN_OFFSET];

			//save record event trigger
			selRecord->event_trigger= responseData[RECORD_BASE + RET_OFFSET];

			//save record event data
			selRecord->event_data1	= responseData[RECORD_BASE + RDATA_OFFSET];
			selRecord->event_data2	= responseData[RECORD_BASE + RDATA_OFFSET+1];
			selRecord->event_data3	= responseData[RECORD_BASE + RDATA_OFFSET+2];

			if (inRecordID == 0 || inRecordID == selRecord->record_id)
			{
				/* We return success if the input record is 
				   begin-of-SEL (value 0), or             
                                   input and output record id matches.   
				*/
				return 0;
			}
			else 
			{
				/* If last record, inRecordID will be -1 from
				   response data, so return -1 as normal EOF. 
				   (fix to bug 279)
				*/
				if (inRecordID == -1) return(-1);
				/* If not last record, this is an error. */
				if (fdebug)
				SELprintf("GetSelEntry: input id  %d != output id %d \n", inRecordID, selRecord->record_id);
				*pRecordID = inRecordID;  //restore the input record id
				return -1;
			}
		}
	}

	// we are here because after the retry, completionCode is not COMPLETION_CODE_OK

	if (fdebug) 
		SELprintf("GetSelEntry: sendImbRequest error completion code, code=%d\n", completionCode);
 	return -1;
 
}  /* end GetSelEntry() */

void StartWriting(uint *plasttime, short *plastid)
{
	FILE *fd;
	uint lasttime; 
	short lastid;
	int ret = -1;
	
	lasttime = 0;
	lastid = 0;
	// Open the index file
	fd = fopen(idxfile,"r");
	if (fd != NULL) {
		// Read the file, get savtime & savid
		ret = fscanf(fd,"%x %x",&lasttime,&lastid);
		fclose(fd);
	}
	if (fdebug) printf("StartWriting: ret = %d, savtime = %x, savid = %x\n",
				ret,lasttime,(ushort)lastid);
	*plasttime = lasttime;
	*plastid = lastid;
	// Open syslog
        openlog( "SEL: ",  LOG_CONS, LOG_KERN);
	return;
}

void StopWriting(uint lasttime, short lastid)
{
	FILE *fd;
	// Rewrite the saved time & record id
	fd = fopen(idxfile,"w");
	if (fd != NULL) {
		fprintf(fd,"%x %x\n",lasttime,lastid);
		fclose(fd);
		}
	// Close syslog
	closelog();
	return;
}

void ClearSEL(void)
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar inputData[6];
	uchar completionCode;
	ushort cmd;

	cmd = RESERVE_SEL;
	status = ipmi_cmd(cmd, inputData, 0, responseData, 
			&responseLength, &completionCode, fdebug);

	if (status == ACCESS_OK && completionCode == 0) {
	    cmd = CLEAR_SEL;
	    inputData[0] = responseData[0];  /*Reservation ID [0]*/
	    inputData[1] = responseData[1];  /*Reservation ID [1]*/
	    inputData[2] = 'C';
	    inputData[3] = 'L';
	    inputData[4] = 'R';
	    inputData[5] = 0xAA;   /* initiate erase */
	    status = ipmi_cmd(cmd, inputData, 6, responseData, 
	  		      &responseLength, &completionCode, fdebug);
	    /* The reservation is cancelled by the CLEAR_SEL cmd */
	    }
	if (status == ACCESS_OK) {
		if( completionCode ) {
		    SELprintf("ClearSEL: cmd = %x, completion code=%x\n",
			      cmd,completionCode); 
		} else {
		    /* Successful, done. */
		    SELprintf("ClearSEL: Log Cleared successfully\n"); 
		}
	} else {
	    SELprintf("ClearSEL: cmd = %x, error status = %x\n",
		      cmd,status); 
	}
	if (fwritesel) {  /* this won't always be enough, but try */
	   StartWriting(&savtime,&savid);
	   savid = 0;  
	   StopWriting(savtime,savid);
	}
}  /* ClearSEL()*/

void ReadSEL(uchar mytype, char fwriteit)
{
	short RecordID = 0;  /* 0 = first record, -1 = end */
	SEL_RECORD selRecord;
	SEL_RECORD *pSelRecord = &selRecord;
	uchar dtype;
	char output[132];
	char mystr[26] = "(123)";  /*for panic string*/
	char *pstr;
	char genstr[10] = "03  ";
	char poststr[24] = "OEM Post Code = %x%x";
	char *gstr;
	int i, j;
	int rc = 0;
	time_t eventTime;
	uchar *evtime;
	uchar timebuf[40];
	
	memset(pSelRecord, 0, sizeof(SEL_RECORD));
	if (fwriteit) { 
		StartWriting(&savtime,&savid);
		RecordID = savid;
	}
	while( rc == 0 ) {
		rc = GetSelEntry( &RecordID, pSelRecord);
		if (fwriteit && (rc != 0) && (RecordID == savid)) {
		    /* If here, log was probably cleared, so try
		     * again from the log start. */
		    RecordID = 0;
		    rc = GetSelEntry( &RecordID, pSelRecord);
		}
		if (rc != 0) { /* EOF or error */ break; }
		if (pSelRecord->record_type >= 0xe0) { /*OEM Record 26.3*/
		   char *pc;
		   pc = (char *)&pSelRecord->timestamp;  /*bytes 4:16*/
		   sprintf(output,"%04x OEM Event %02x %02x ", 
			(ushort)pSelRecord->record_id, pc[0], pc[1]);
		   j = strlen(output);
		   for (i = 2; i < 13; i++) {  /* 4:16 = 13 bytes data */
		      if (pSelRecord->record_type == 0xf0) { 
		        /* panic string will be type 0xf0 */
			if (pc[i] == 0) break;
			output[j++] = pc[i];
		      } else {
			sprintf(&output[j],"%02x ",pc[i]);
			j += 3;
		      }
		   }
		   output[j++] = '\n';
		   output[j++] = 0;
		   printf(output);
		   if (fwriteit) {
			/* Only write newer SEL records to syslog */
			if (pSelRecord->record_id > savid) {
			   syslog(LOG_INFO,output);
			   savid = pSelRecord->record_id;
			}
		   }
		} else if (pSelRecord->record_type == 0x02) {
		  /* most records are type 2 */
		  if (mytype == 0xff || pSelRecord->sensor_type == mytype) {
		    /* set dtype, used as array index below */
		    if (pSelRecord->sensor_type > 0x29) dtype = 0;
		    else dtype = pSelRecord->sensor_type;
		    if (pSelRecord->sensor_type == 0x20) { /*OS Crit Stop*/
		        /* Show first 3 chars of panic string */
		        mystr[0] = '(';
		        mystr[1] = pSelRecord->sensor_number & 0x7f;
		        mystr[2] = pSelRecord->event_data2 & 0x7f;
		        mystr[3] = pSelRecord->event_data3 & 0x7f;
			mystr[4] = ')';
			mystr[5] = 0;
			pstr = mystr;
			if (pSelRecord->sensor_number & 0x80) 
				strcat(mystr,"Oops!");
			if (pSelRecord->event_data2 & 0x80) 
				strcat(mystr,"Int!");
			if (pSelRecord->event_data3 & 0x80) 
				strcat(mystr,"NullPtr!");
		    } else if (pSelRecord->sensor_type == 0x0f) { /*Post Errs*/
			pstr = "POST Code";
			if (pSelRecord->event_data1 == 0xa0) {
			   /* OEM post codes in bytes 2 & 3 (lo-hi) */
			   sprintf(poststr,"POST Code %02x%02x",
      			        pSelRecord->event_data3,
      			        pSelRecord->event_data2);
			   pstr = poststr;
			} else
			for (i=0; posterrs[i].code0 != 0x00; i++) {
      			   if ((posterrs[i].code0==pSelRecord->event_data1) &&
          		      (posterrs[i].code1==pSelRecord->event_data2)) {
        		     pstr = posterrs[i].msg;
        		     break;
        		  }
      			} /*end for*/
		    } else {
			mystr[0] = '-'; mystr[1] = 0;
			pstr = mystr;
			/* Use sens_desc array for other misc descriptions */
			for (i = 0; i < NSDESC; i++) {
			   if (sens_desc[i].s_typ == pSelRecord->sensor_type) {
			      if (sens_desc[i].s_num != 0xff &&
			         sens_desc[i].s_num !=pSelRecord->sensor_number)
				    continue;
			      if (sens_desc[i].evtrg != 0xff &&
			         sens_desc[i].evtrg !=pSelRecord->event_trigger)
				    continue;
			      if (sens_desc[i].data1 != 0xff && 
      			          sens_desc[i].data1 != pSelRecord->event_data1)
				    continue;
			      if (sens_desc[i].data2 != 0xff && 
      			          sens_desc[i].data2 != pSelRecord->event_data2)
				    continue;
			      if (sens_desc[i].data3 != 0xff && 
      			          sens_desc[i].data3 != pSelRecord->event_data3)
				    continue;
			      /* have a match, use description */
			      pstr = (char *)sens_desc[i].desc; 
			      break;
		 	   }
      			} /*end for*/
		    }
		    /* show the Generator ID */
		    sprintf(genstr,"%02x  ",pSelRecord->generator_id);
		    gstr = genstr;  /* default */
		    for (i = 0; i < NGDESC; i++)
		    {
			if (gen_desc[i].g_id == pSelRecord->generator_id)
			   gstr = (char *)gen_desc[i].desc;  
		    }

		    evtime = (uchar *)&pSelRecord->timestamp;
		    eventTime = evtime[0] + (evtime[1] << 8) + 
		   		(evtime[2] << 16) + (evtime[3] << 24);
		    strftime(timebuf,sizeof(timebuf),
                             "%x %H:%M:%S",localtime(&eventTime));
		    //  printf("%s\n", timebuf);
		    sprintf(output,
			"%04x %s "
			"%s %02x %s %02x %s %02x [%02x %02x %02x]\n",
			(ushort)pSelRecord->record_id,
			timebuf,     /*  pSelRecord->timestamp */
			/* record_type,  2 = SEL record */
			/* evm_rev,   event msg rev =3 if 1.0, =4 if 1.5 */
			gstr,   /* pSelRecord->generator_id */
			pSelRecord->sensor_type,
			sensor_types[dtype] ,
			pSelRecord->sensor_number,
			pstr,
			pSelRecord->event_trigger,
			pSelRecord->event_data1,
			pSelRecord->event_data2,
			pSelRecord->event_data3 );
		    printf(output);
		    if (fwriteit) {
			/* Only write newer records */
			if (pSelRecord->timestamp > savtime) {
			   /* Write the given SEL record to the syslog.  */
			   syslog(LOG_INFO,output);
			   savtime = pSelRecord->timestamp;
			   savid = pSelRecord->record_id;
			}
		    } /*endif fwriteit*/
		 }  /*endif match*/
		}  /*endif type 2 */
		else {  /* other misc record type */
		   char *pc;
		   pc = (char *)&pSelRecord->record_type;
		   printf(output,"%04x Type%02x ", 
			(ushort)pSelRecord->record_id,pc[0]);
		   j = strlen(output);
		   for (i = 1; i < 14; i++) {
			sprintf(genstr,"%02x ",pc[i]);
			strcat(output,genstr);
		   }
		   strcat(output,"\n");
		   printf(output);
		   if (fwriteit) {
			/* Only write newer records  to syslog */
			if (pSelRecord->record_id > savid) {
			   syslog(LOG_INFO,output);
			   savid = pSelRecord->record_id;
			}
		   }
		}  /*endif misc tupe*/
		if( pSelRecord->record_id == -1 )
			break;
		if( RecordID == pSelRecord->record_id )
			break;
		memset(pSelRecord, 0, sizeof(SEL_RECORD));
	}  /*endwhile*/
	if (fwriteit) StopWriting(savtime,savid);
}  /* end ReadSEL()*/

ushort vfree;

static int ReadSELinfo()
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar completionCode;
	uchar inputData[6];

	status = ipmi_cmd(GET_SEL_INFO, inputData, 0, responseData, 
			&responseLength, &completionCode, fdebug);
	if (fdebug) printf("ipmi_cmd status = %x\n",status);

	if (status == ACCESS_OK) {
		vfree = responseData[3] + (responseData[4] * 0x100);
		SELprintf("Code %d SEL Ver %d Support %d, free space %d\n",
		          completionCode, 
		          responseData[0],
		          responseData[13],
			  vfree  );
		//successful, done
		return(0);
	} else    return(1);

}  /*end ReadSELinfo()*/

int
main(int argc, char **argv)
{
   int ret;
   char c;
   char DevRecord[16];

   printf("%s: version %s\n",progname,progver);
   while ( (c = getopt( argc, argv,"cswvx?")) != EOF )
      switch(c) {
          case 'x': fdebug = 1;      break;
          case 'w': fwritesel = 1;   break;
          case 's': fall = 0;        break;
          case 'c': fclearsel = 1;   break;
          case 'v': fonlyver = 1;    break;
          default:
                printf("Usage: %s [-cswx]\n", progname);
		printf("   -c  Clears all SEL records\n");
		printf("   -s  Show only Critical Stop records\n");
		printf("   -w  Writes new SEL records to syslog\n");
		printf("   -v  Only show version information\n");
		printf("   -x  Display extra debug messages\n");
                exit(1);
      }               
   // if (fwrite) fall = 1;   // would we ever set fwrite without fall??

   ret = ipmi_getdeviceid( DevRecord, sizeof(DevRecord));
   if (ret == 0) {
      uchar ipmi_maj, ipmi_min;
      ipmi_maj = DevRecord[4] & 0x0f;
      ipmi_min = DevRecord[4] >> 4;
      printf("-- BMC version %x.%x, IPMI version %d.%d \n",
             DevRecord[2],  DevRecord[3],
             ipmi_maj, ipmi_min);
      }
   ret = ReadSELinfo();
   if (ret == 0 && !fonlyver) {
      if (fclearsel) {
	 ClearSEL();
      } else {
         printf("%s: starting ReadSEL ...\n",progname);
         if (fall) ReadSEL(0xff,fwritesel); /* show all SEL records */
         else ReadSEL(0x20,fwritesel);      /* only show OS Crit Stops*/
	 if (vfree < MIN_FREE) {
	     /* PEF alerts and other log messages fail if low free space,
		so show a warning. */
             printf("WARNING: free space is low (=%d), need to clear with -c\n",
		    vfree);
	 }
      }
   }
   exit(ret);
}

/* end showsel.c */
